@@ -1,6 +1,6 @@
 #!/bin/bash
 #
-# Copyright (C) 2008-2021 Michele Martone
+# Copyright (C) 2008-2022 Michele Martone
 # 
 # This file is part of librsb.
 # 
@@ -49,6 +49,7 @@ if which rats; then
 	echo "output of running rats in rats.log"
 fi
 if cat *.F90 examples/*.F90| sed 's/!.*$//g'| grep '^.\{73,\}' ; then echo 'Some source code exceeds 72 chars!'; fi
+if grep 'x"ac_cv_' configure.ac */configure.ac; then false ; else true; fi
 if cat examples/*.cpp| grep '^.\{155,\}' ; then echo 'Some source code reaches 155 chars!'; fi
 test -f NEWS
 test -f README
@@ -63,6 +64,16 @@ if grep '	' *.F90 */*.F90 ; then exit 255 ; else true ; fi
 for f in  examples/*0 examples/*c   ;
 	do grep -q  -F  $f rsb.h || echo "$f is missing from among rsb.h mentioned examples!";
 done
+
+OUT=`./librsb-config --I_opts`
+test `echo "$OUT" |  wc -l` -le 1
+OUT=`./librsb-config --cxxflags`
+test `echo "$OUT" |  wc -l` -le 1
+OUT=`./librsb-config --ldflags`
+test `echo "$OUT" |  wc -l` -le 1
+OUT=`./librsb-config --libs --extra_libs`
+test `echo "$OUT" |  wc -l` -le 1
+
 ./rsbench -E 0.1s || exit 255
 #
 if test -z "${RSB_DEVTESTS_NO_LIMITS_TEST}" ; then
